#include <math.h>
#include <string.h>
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "colorspace.h"
#include "subsample.h"

#include <yuv4mpeg.h>

#if defined(HAVE_PDL)
  #include <pdlcore.h>

  static Core* PDL;
  static SV* CoreSV;
#endif

typedef y4m_stream_info_t *YUV4MPEG__StreamInfo;
typedef y4m_frame_info_t  *YUV4MPEG__FrameInfo;
typedef y4m_ratio_t *YUV4MPEG__Ratio;
typedef y4m_xtag_list_t *YUV4MPEG__XTagList;

typedef PerlIO *InputHandle;
typedef PerlIO *OutputHandle;

typedef struct {
  y4m_stream_info_t stream;
  int w;
  int h;
  int count;
  int interlace;
  int deinterlaced;

  uint8_t *top[3];
  uint8_t *bottom[3];

  uint8_t *data[3];
} frame_t;

typedef frame_t *YUV4MPEG__Frame;

void deinterlace_frame(YUV4MPEG__Frame frame)
{
  int x;
  int i = 0, j = 0;
  int w = frame->w;

  while (j < frame->count)
    {
       for (x = 0; x < w; x++, i++, j++)
         {
           frame->data[0][j] = frame->top[0][i];
           frame->data[1][j] = frame->top[1][i];
           frame->data[2][j] = frame->top[2][i];

           frame->data[0][j + w] = frame->bottom[0][i];
           frame->data[1][j + w] = frame->bottom[1][i];
           frame->data[2][j + w] = frame->bottom[2][i];
         }

       j += w;
    }
}

void interlace_frame(YUV4MPEG__Frame frame)
{
  int x;
  int i = 0, j = 0;
  int w = frame->w;

  while (j < frame->count)
    {
       for (x = 0; x < w; x++, i++, j++)
         {
           frame->top[0][i] = frame->data[0][j];
           frame->top[1][i] = frame->data[1][j];
           frame->top[2][i] = frame->data[2][j];

           frame->bottom[0][i] = frame->data[0][j + w];
           frame->bottom[1][i] = frame->data[1][j + w];
           frame->bottom[2][i] = frame->data[2][j + w];
         }

       j += w;
    }
}

#if defined(HAVE_PDL)

static pdl* new_pdl(int datatype, PDL_Long dims[], int ndims)
{
  pdl *p = PDL->pdlnew();
  PDL->setdims (p, dims, ndims);  /* set dims */
  p->datatype = datatype;         /* and data type */
  PDL->allocdata (p);             /* allocate the data chunk */

  return p;
}

#endif

MODULE = YUV4MPEG       PACKAGE = YUV4MPEG::StreamInfo      PREFIX = y4m_si_

PROTOTYPES: DISABLE

YUV4MPEG::StreamInfo new(package, source=NULL)
        char *package;
        YUV4MPEG::StreamInfo source;
        CODE:
        New(42, RETVAL, 1, y4m_stream_info_t);
        y4m_init_stream_info(RETVAL);
        
        if (source != NULL)
        {
          y4m_copy_stream_info(RETVAL, source);
        }

        OUTPUT:
        RETVAL

void DESTROY(stream)
        YUV4MPEG::StreamInfo stream;
        CODE:
        y4m_fini_stream_info(stream);
        Safefree(stream);

void y4m_si_set_width(stream, width)
        YUV4MPEG::StreamInfo stream;
        int width;

int y4m_si_get_width(stream)
        YUV4MPEG::StreamInfo stream;

void y4m_si_set_height(stream, height)
        YUV4MPEG::StreamInfo stream;
        int height;

int y4m_si_get_height(stream)
        YUV4MPEG::StreamInfo stream;

void y4m_si_set_interlace(stream, interlace)
        YUV4MPEG::StreamInfo stream;
        int interlace;

int y4m_si_get_interlace(stream)
        YUV4MPEG::StreamInfo stream;

void y4m_si_set_framerate(stream, ratio)
        YUV4MPEG::StreamInfo stream;
        YUV4MPEG::Ratio ratio;
        CODE:
        y4m_si_set_framerate(stream, *ratio);

YUV4MPEG::Ratio y4m_si_get_framerate(stream)
        YUV4MPEG::StreamInfo stream;
        CODE:
        New(42, RETVAL, 1, y4m_ratio_t);
        *RETVAL = y4m_si_get_framerate(stream);
        OUTPUT:
        RETVAL

void y4m_si_set_sampleaspect(stream, ratio)
        YUV4MPEG::StreamInfo stream;
        YUV4MPEG::Ratio ratio;
        CODE:
        y4m_si_set_sampleaspect(stream, *ratio);

YUV4MPEG::Ratio y4m_si_get_sampleaspect(stream)
        YUV4MPEG::StreamInfo stream;
        CODE:
        New(42, RETVAL, 1, y4m_ratio_t);
        *RETVAL = y4m_si_get_sampleaspect(stream);
        OUTPUT:
        RETVAL

int y4m_si_get_framelength(stream)
        YUV4MPEG::StreamInfo stream;

YUV4MPEG::XTagList y4m_si_xtags(stream)
        YUV4MPEG::StreamInfo stream;

int parse(stream, tags)
        YUV4MPEG::StreamInfo stream;
        SV *tags;
        
        PREINIT:
        char *buf;      
        int len;

        CODE:
        buf = SvPV(tags, len);
        RETVAL = y4m_parse_stream_tags(buf, stream);

        OUTPUT:
        RETVAL

int read(stream, fd=PerlIO_stdin())
        YUV4MPEG::StreamInfo stream;
        InputHandle fd;
        CODE:
        RETVAL = y4m_read_stream_header(PerlIO_fileno(fd), stream);
        OUTPUT:
        RETVAL

int write(stream, fd=PerlIO_stdout())
        YUV4MPEG::StreamInfo stream;
        OutputHandle fd;
        CODE:
        RETVAL = y4m_write_stream_header(PerlIO_fileno(fd), stream);
        OUTPUT:
        RETVAL        

MODULE = YUV4MPEG       PACKAGE = YUV4MPEG::FrameInfo      PREFIX = y4m_fi_

YUV4MPEG::FrameInfo new(package, source=NULL)        
        char *package;
        YUV4MPEG::FrameInfo source;
        CODE:
        New(42, RETVAL, 1, y4m_frame_info_t);
        y4m_init_frame_info(RETVAL);
        
        if (source != NULL)
        {
          y4m_copy_frame_info(RETVAL, source);
        }

        OUTPUT:
        RETVAL

void DESTROY(frame)
        YUV4MPEG::FrameInfo frame;
        CODE:
        y4m_fini_frame_info(frame);
        Safefree(frame);

#if defined(HAVE_Y4M_FI_XTAGS)

YUV4MPEG::XTagList y4m_fi_xtags(stream)
        YUV4MPEG::FrameInfo stream;

#else

void y4m_fi_xtags(stream)
        YUV4MPEG::FrameInfo stream;
        
        PPCODE:
        croak("This mjpegutils is missing support for frame tags");

#endif

int read(frame, fd=PerlIO_stdin())
        YUV4MPEG::FrameInfo frame;
        InputHandle fd;
        CODE:
        RETVAL = y4m_read_frame_header(PerlIO_fileno(fd), frame);
        OUTPUT:
        RETVAL

int write(frame, fd=PerlIO_stdout())
        YUV4MPEG::FrameInfo frame;
        OutputHandle fd;
        CODE:
        RETVAL = y4m_write_frame_header(PerlIO_fileno(fd), frame);
        OUTPUT:
        RETVAL

MODULE = YUV4MPEG       PACKAGE = YUV4MPEG::Frame

YUV4MPEG::Frame new(package, stream)
        char *package;
        YUV4MPEG::StreamInfo stream;
        CODE:
        New(42, RETVAL, 1, frame_t);

        y4m_copy_stream_info(&(RETVAL->stream), stream);        
        RETVAL->w = y4m_si_get_width(stream);
        RETVAL->h = y4m_si_get_height(stream);
        RETVAL->interlace = y4m_si_get_interlace(stream);
        RETVAL->count = RETVAL->w * RETVAL->h;
        RETVAL->deinterlaced = 0;
        
        Newz(42, RETVAL->data[0], RETVAL->count, uint8_t);
        Newz(42, RETVAL->data[1], RETVAL->count, uint8_t);
        Newz(42, RETVAL->data[2], RETVAL->count, uint8_t);

        if (RETVAL->interlace != Y4M_ILACE_NONE)
        {
          Newz(42, RETVAL->top[0], RETVAL->count / 2, uint8_t);
          Newz(42, RETVAL->top[1], RETVAL->count / 2, uint8_t);
          Newz(42, RETVAL->top[2], RETVAL->count / 2, uint8_t);

          Newz(42, RETVAL->bottom[0], RETVAL->count / 2, uint8_t);
          Newz(42, RETVAL->bottom[1], RETVAL->count / 2, uint8_t);
          Newz(42, RETVAL->bottom[2], RETVAL->count / 2, uint8_t);
        }

        OUTPUT:
        RETVAL

void DESTROY(frame)
        YUV4MPEG::Frame frame

        CODE:
        Safefree(frame->data[0]);
        Safefree(frame->data[1]);
        Safefree(frame->data[2]);

        if (frame->interlace != Y4M_ILACE_NONE)
        {
          Safefree(frame->top[0]);
          Safefree(frame->top[1]);
          Safefree(frame->top[2]);

          Safefree(frame->bottom[0]);
          Safefree(frame->bottom[1]);
          Safefree(frame->bottom[2]);
        }

        Safefree(frame);

int read(frame, fi, fd=PerlIO_stdin())
        YUV4MPEG::Frame frame
        InputHandle fd;
        YUV4MPEG::FrameInfo fi;

        CODE:        
        switch(frame->interlace) 
        {
        case Y4M_ILACE_NONE:
          RETVAL = y4m_read_frame(PerlIO_fileno(fd), &(frame->stream), fi, 
                                  frame->data);          
          frame->deinterlaced = 1;
          break;

        case Y4M_ILACE_TOP_FIRST:
          RETVAL = y4m_read_fields(PerlIO_fileno(fd), &(frame->stream), fi,
                                   frame->top, frame->bottom);
          break;

        case Y4M_ILACE_BOTTOM_FIRST:
          RETVAL = y4m_read_fields(PerlIO_fileno(fd), &(frame->stream), fi,
                                   frame->bottom, frame->top);
          break;

        default:
          croak("unknown interlacing mode");
        }

        OUTPUT:
        RETVAL

int write(frame, fi, fd=PerlIO_stdout())
        YUV4MPEG::Frame frame
        OutputHandle fd;
        YUV4MPEG::FrameInfo fi;

        CODE:        
        switch(frame->interlace) 
        {
        case Y4M_ILACE_NONE:
          RETVAL = y4m_write_frame(PerlIO_fileno(fd), &(frame->stream), fi, 
                                   frame->data);
          break;

        case Y4M_ILACE_TOP_FIRST:
          interlace_frame(frame);

          RETVAL = y4m_write_fields(PerlIO_fileno(fd), &(frame->stream), fi,
                                    frame->top, frame->bottom);
          break;

        case Y4M_ILACE_BOTTOM_FIRST:
          interlace_frame(frame);

          RETVAL = y4m_write_fields(PerlIO_fileno(fd), &(frame->stream), fi,
                                    frame->bottom, frame->top);
          break;

        default:
          croak("unknown interlacing mode");
        }

        OUTPUT:
        RETVAL

void get_data(frame)
        YUV4MPEG::Frame frame

        PREINIT:
        AV* p1 = newAV();
        AV* p2 = newAV();
        AV* p3 = newAV();

        SV* r1;
        SV* r2;
        SV* r3;

        int i;
                
        PPCODE:
        if (! frame->deinterlaced)
        {
          deinterlace_frame(frame);
        }

        av_extend(p1, frame->count);        
        av_extend(p2, frame->count);        
        av_extend(p3, frame->count);

        for (i = 0; i < frame->count; i++)
        {
          av_store(p1, i, newSViv(frame->data[0][i]));
          av_store(p2, i, newSViv(frame->data[1][i]));
          av_store(p3, i, newSViv(frame->data[2][i]));
        }

        r1 = newRV(sv_2mortal((SV *)p1));
        r2 = newRV(sv_2mortal((SV *)p2));
        r3 = newRV(sv_2mortal((SV *)p3));

        XPUSHs(sv_2mortal(r1));
        XPUSHs(sv_2mortal(r2));
        XPUSHs(sv_2mortal(r3));

void set_data(frame, p1, p2, p3)
        YUV4MPEG::Frame frame;
        AV *p1;
        AV *p2;
        AV *p3;

        PREINIT:
        int i;

        CODE:
        frame->deinterlaced = 1;

        for (i = 0; i < frame->count; i++)
        {
          frame->data[0][i] = SvIV(*av_fetch(p1, i, 0));
          frame->data[1][i] = SvIV(*av_fetch(p2, i, 0));
          frame->data[2][i] = SvIV(*av_fetch(p3, i, 0));
        }

void get_combined_data(frame)        
        YUV4MPEG::Frame frame
        PREINIT:
        AV* data = newAV();
        SV* ref;
        int i, j;
                
        PPCODE:
        if (! frame->deinterlaced)
        {
          deinterlace_frame(frame);
        }

        av_extend(data, frame->count * 3);

        for (i = 0, j = 0; i < frame->count; i++, j += 3)
        {
          av_store(data, j, newSViv(frame->data[0][i]));
          av_store(data, j + 1, newSViv(frame->data[1][i]));
          av_store(data, j + 2, newSViv(frame->data[2][i]));
        }

        ref = newRV(sv_2mortal((SV *)data));

        XPUSHs(sv_2mortal(ref));

void set_combined_data(frame, data)
        YUV4MPEG::Frame frame;
        AV *data;

        PREINIT:
        int i, j;

        CODE:
        frame->deinterlaced = 1;

        for (i = 0, j = 0; i < frame->count; i++, j += 3)
        {
          frame->data[0][i] = SvIV(*av_fetch(data, j, 0));
          frame->data[1][i] = SvIV(*av_fetch(data, j + 1, 0));
          frame->data[2][i] = SvIV(*av_fetch(data, j + 2, 0));
        }

void get_packed_data(frame)        
        YUV4MPEG::Frame frame
        PREINIT:
        char *data;
        int i, j;
                
        PPCODE:
        if (! frame->deinterlaced)
        {
          deinterlace_frame(frame);
        }

        New(42, data, frame->count * 3, char);
        SAVEFREEPV(data);

        for (i = 0, j = 0; i < frame->count; i++, j += 3)
        {
          data[j]     = frame->data[0][i];
          data[j + 1] = frame->data[1][i];
          data[j + 2] = frame->data[2][i];
        }

        XPUSHs(sv_2mortal(newSVpv(data, frame->count * 3)));

void set_packed_data(frame, data_sv)
        YUV4MPEG::Frame frame;
        SV *data_sv;

        PREINIT:
        int i, j;
        char *data;
        int len;

        CODE:
        frame->deinterlaced = 1;

        data = SvPV(data_sv, len);

        for (i = 0, j = 0; j < len; i++, j += 3)
        {
          frame->data[0][i] = data[j];
          frame->data[1][i] = data[j + 1];
          frame->data[2][i] = data[j + 2];
        }

#if defined(HAVE_PDL)

pdl* get_pdl_data(frame)
        YUV4MPEG::Frame frame
        
        PREINIT:
        PDL_Long dims[3];
        PDL_Byte *data;
        int i, pos, y, base;

        CODE:
        if (! frame->deinterlaced)
        {
          deinterlace_frame(frame);
        }

        dims[0] = 3;
        dims[1] = frame->w;
        dims[2] = frame->h;

        RETVAL = new_pdl(PDL_B, dims, 3);

        data = (PDL_Byte *)RETVAL->data;

        for (y = (frame->h - 1), i = 0; y >= 0; y--)
        { 
          base = y * frame->w;

          for (pos = base; pos < (base + frame->w); pos++, i++)
          {
            data[pos * 3] = frame->data[0][i];
            data[pos * 3 + 1] = frame->data[1][i];
            data[pos * 3 + 2] = frame->data[2][i];
          }
        }

        OUTPUT:
        RETVAL

void set_pdl_data(frame, p)
        YUV4MPEG::Frame frame
        pdl* p;
        
        PREINIT:
        PDL_Byte *data;
        int i, pos, y, base;

        CODE:
        if ((p->ndims != 3) ||
            (p->dims[0] != 3) ||
            (p->dims[1] != frame->w) ||
            (p->dims[2] != frame->h))
        {
          croak("was expecting pdl with dimensions (3,%d,%d)",
                frame->w, frame->h);
        }

        frame->deinterlaced = 1;

        data = (PDL_Byte *)p->data;

        for (y = (frame->h - 1), i = 0; y >= 0; y--)
        { 
          base = y * frame->w;

          for (pos = base; pos < (base + frame->w); pos++, i++)
          {
            frame->data[0][i] = data[pos * 3];
            frame->data[1][i] = data[pos * 3 + 1];
            frame->data[2][i] = data[pos * 3 + 2];
          }
        }

#else

void get_pdl_data(frame)
        YUV4MPEG::Frame frame;

        PPCODE:
        croak("PDL is not installed");

void set_pdl_data(frame, p)
        YUV4MPEG::Frame frame;
        SV *p;

        PPCODE:
        croak("PDL is not installed");

#endif

void to_rgb(frame, mode=SSM_420_JPEG)
        YUV4MPEG::Frame frame;
        int mode;
        
        CODE:
        chroma_supersample(mode, frame->data, frame->w, frame->h);
        convert_YCbCr_to_RGB(frame->data, frame->count);
        
void to_yuv(frame, mode=SSM_420_JPEG)
        YUV4MPEG::Frame frame;
        int mode;

        CODE:
        convert_RGB_to_YCbCr(frame->data, frame->count);
        chroma_subsample(mode, frame->data, frame->w, frame->h);

void chroma_supersample(frame, mode=SSM_420_JPEG)
        YUV4MPEG::Frame frame;
        int mode;

        CODE:
        chroma_supersample(mode, frame->data, frame->w, frame->h);

void chroma_subsample(frame, mode=SSM_420_JPEG)
        YUV4MPEG::Frame frame;
        int mode;

        CODE:        
        chroma_subsample(mode, frame->data, frame->w, frame->h);

MODULE = YUV4MPEG       PACKAGE = YUV4MPEG::Ratio

YUV4MPEG::Ratio new(package, n=0, d=1)
        char *package;
        int n;
        int d;
        CODE:
        New(42, RETVAL, 1, y4m_ratio_t);
        RETVAL->n = n;
        RETVAL->d = d;
        OUTPUT:
        RETVAL

void DESTROY(ratio)
        YUV4MPEG::Ratio ratio;
        CODE:
        Safefree(ratio);

int get_n(ratio)
        YUV4MPEG::Ratio ratio;
        CODE:
        RETVAL = ratio->n;
        OUTPUT:
        RETVAL

int get_d(ratio)
        YUV4MPEG::Ratio ratio;
        CODE:
        RETVAL = ratio->d;
        OUTPUT:
        RETVAL

void get(ratio)
        YUV4MPEG::Ratio ratio;
        PPCODE:
        XPUSHs(sv_2mortal(newSViv(ratio->n)));
        XPUSHs(sv_2mortal(newSViv(ratio->d)));

void set_n(ratio, n)
        YUV4MPEG::Ratio ratio;  
        int n;
        CODE:
        ratio->n = n;

void set_d(ratio, d)
        YUV4MPEG::Ratio ratio;  
        int d;
        CODE:

        if (d == 0) 
        {
          croak("can't have zero as a ratio denominator");
        }

        ratio->d = d;

void set(ratio, n, d)
        YUV4MPEG::Ratio ratio;
        int n;
        int d;
        CODE:

        if (d == 0) 
        {
          croak("can't have zero as a ratio denominator");
        }

        ratio->n = n;
        ratio->d = d;

void reduce(ratio)
        YUV4MPEG::Ratio ratio;
        CODE:
        y4m_ratio_reduce(ratio);

int parse(ratio, str)
        YUV4MPEG::Ratio ratio;
        SV *str;

        PREINIT:
        char *strbuf;
        int len;
       
        CODE:
        strbuf = SvPV(str, len);
        RETVAL = y4m_parse_ratio(ratio, strbuf);

        OUTPUT:
        RETVAL

int equals(ratio, ratio2)
        YUV4MPEG::Ratio ratio;
        YUV4MPEG::Ratio ratio2;
        CODE:
        RETVAL = (ratio->n == ratio2->n) &&
                 (ratio->d == ratio2->d);
        OUTPUT:
        RETVAL

YUV4MPEG::Ratio guess_sar(dar, w, h)
        YUV4MPEG::Ratio dar;
        int w;
        int h;
        
        CODE:
        New(42, RETVAL, 1, y4m_ratio_t);

        *RETVAL = y4m_guess_sar(w, h, *dar);

        OUTPUT:
        RETVAL

MODULE = YUV4MPEG       PACKAGE = YUV4MPEG::XTagList      PREFIX = y4m_xtag_

void copy(dest, src)
        YUV4MPEG::XTagList dest;
        YUV4MPEG::XTagList src;

        CODE:
        y4m_copy_xtag_list(dest, src);

int y4m_xtag_count(list)
        YUV4MPEG::XTagList list;

char *y4m_xtag_get(list, n)
        YUV4MPEG::XTagList list;
        int n;

int add(list, tag)
        YUV4MPEG::XTagList list;
        SV *tag;

        PREINIT:
        char *buf;
        int len;

        CODE:
        buf = SvPV(tag, len);
        RETVAL = y4m_xtag_add(list, buf);

        OUTPUT:
        RETVAL

int y4m_xtag_remove(list, n)
        YUV4MPEG::XTagList list;
        int n;

int y4m_xtag_clearlist(list)
        YUV4MPEG::XTagList list;

int y4m_xtag_addlist(list, src)
        YUV4MPEG::XTagList list;
        YUV4MPEG::XTagList src;

MODULE = YUV4MPEG               PACKAGE = YUV4MPEG      PREFIX = y4m_

BOOT:       
#if defined(HAVE_PDL)
   perl_require_pv ("PDL::Core"); /* make sure PDL::Core is loaded */
   CoreSV = perl_get_sv("PDL::SHARE",FALSE);  /* SV* value */
#ifndef aTHX_
#define aTHX_
#endif
   if (CoreSV==NULL)
     Perl_croak(aTHX_ "We require the PDL::Core module, which was not found");
   PDL = INT2PTR(Core*,SvIV( CoreSV ));  /* Core* value */
   if (PDL->Version != PDL_CORE_VERSION)
     Perl_croak(aTHX_ "The code needs to be recompiled against the newly installed PDL");
#endif

char *y4m_strerr(n)
        int n;

int y4m_allow_unknown_tags(n)
        int n;

YUV4MPEG::Ratio _load_ratio_constant(ref)
        int ref

        PREINIT:
        y4m_ratio_t *ratio;

        CODE:
        ratio = (y4m_ratio_t *)ref;

        New(42, RETVAL, 1, y4m_ratio_t);
        RETVAL->n = ratio->n;
        RETVAL->d = ratio->d;

        OUTPUT:
        RETVAL
